package com.sleepycat.je.log;

import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;

import com.sleepycat.je.DatabaseException;
import com.sleepycat.je.dbi.EnvironmentImpl;
import com.sleepycat.je.log.FileManager;
import com.sleepycat.je.config.EnvironmentParams;

public class ChunkedNIOFileManager extends FileManager {

	static aspect ChunkedNIOAspect {

		FileManager around(EnvironmentImpl envImpl, File dbEnvHome,
				boolean readOnly) throws DatabaseException :
			call(FileManager.new(EnvironmentImpl, File, boolean)) && args(envImpl, dbEnvHome, readOnly){

			return new ChunkedNIOFileManager(envImpl, dbEnvHome, readOnly);
		}
	}

	/*
	 * If non-0, do NIO in chunks of this size.
	 */
	private long chunkedNIOSize = 0;

	public ChunkedNIOFileManager(EnvironmentImpl envImpl, File dbEnvHome,
			boolean readOnly) throws DatabaseException {
		
		super(envImpl, dbEnvHome, readOnly);
		
		chunkedNIOSize = envImpl.getConfigManager()
				.getLong(EnvironmentParams.LOG_CHUNKED_NIO);

		if (chunkedNIOSize <= 0) {
			throw new IllegalArgumentException(EnvironmentParams.LOG_CHUNKED_NIO
					.getName()
					+ " must be set to a value > 0 when using the chunked NIO feature.");
		}
	}

	void readFromFile(RandomAccessFile file, ByteBuffer readBuffer, long offset)
			throws IOException {

		FileChannel channel = file.getChannel();

		/*
		 * Read a chunk at a time to prevent large direct memory allocations by
		 * NIO.
		 */
		int readLength = readBuffer.limit();
		long currentPosition = offset;
		while (readBuffer.position() < readLength) {
			readBuffer.limit((int) (Math.min(readBuffer.limit()
					+ chunkedNIOSize, readLength)));
			int bytesRead = channel.read(readBuffer, currentPosition);

			if (bytesRead < 1)
				break;

			currentPosition += bytesRead;
		}
	}

	protected int writeToFile(RandomAccessFile file, ByteBuffer data,
			long destOffset) throws IOException, DatabaseException {

		int totalBytesWritten = 0;
		FileChannel channel = file.getChannel();

		/*
		 * We can't change the limit without impacting readers that might find
		 * this buffer in the buffer pool. Duplicate the buffer so we can set
		 * the limit independently.
		 */
		ByteBuffer useData = data.duplicate();

		/*
		 * Write small chunks of data by manipulating the position and limit
		 * properties of the buffer, and submitting it for writing repeatedly.
		 * 
		 * For each chunk, the limit is set to the position + chunkedNIOSize,
		 * capped by the original limit of the buffer.
		 * 
		 * Preconditions: data to be written is betweek data.position() and
		 * data.limit()
		 * 
		 * Postconditions: data.limit() has not changed, data.position() ==
		 * data.limit(), offset of the channel has not been modified.
		 */
		int originalLimit = useData.limit();
		useData.limit(useData.position());
		while (useData.limit() < originalLimit) {
			useData.limit((int) (Math.min(useData.limit() + chunkedNIOSize,
					originalLimit)));
			int bytesWritten = channel.write(useData, destOffset);
			destOffset += bytesWritten;
			totalBytesWritten += bytesWritten;
		}

		return totalBytesWritten;
	}
}
